import base64
import binascii
import hashlib
import pathlib
from urllib.parse import parse_qs

import httpx


def get_player_id():
    player_id = base64.encodebytes(hashlib.sha1(b"").digest()).rstrip()
    return player_id.decode("ascii")


def extract_token_from_url(url):
    parsed_url = url.query
    query_dict = parse_qs(parsed_url)
    return query_dict["playerToken"]


def get_player_token(auth) -> str:
    with httpx.Client(cookies=auth.website_cookies) as session:
        audible_base_url = f"https://www.audible.{auth.locale.domain}"
        params = {
            "ipRedirectOverride": True,
            "playerType": "software",
            "bp_ua": "y",
            "playerModel": "Desktop",
            "playerId": get_player_id(),
            "playerManufacturer": "Audible",
            "serial": ""
        }
        resp = session.get(f"{audible_base_url}/player-auth-token",
                           params=params)
    
        player_token = extract_token_from_url(resp.url)[0]
    
        return player_token


def extract_activation_bytes(data):
    if (b"BAD_LOGIN" in data or b"Whoops" in data) or \
            b"group_id" not in data:
        print(data)
        print("\nActivation failed! ;(")
        raise ValueError("data wrong")
    a = data.rfind(b"group_id")
    b = data[a:].find(b")")
    keys = data[a + b + 1 + 1:]
    output = []
    output_keys = []
    # each key is of 70 bytes
    for i in range(0, 8):
        key = keys[i * 70 + i:(i + 1) * 70 + i]
        h = binascii.hexlify(bytes(key))
        h = [h[i:i+2] for i in range(0, len(h), 2)]
        h = b','.join(h)
        output_keys.append(h)
        output.append(h.decode("utf-8"))

    # only 4 bytes of output_keys[0] are necessary for decryption! ;)
    activation_bytes = output_keys[0].replace(b",", b"")[0:8]
    # get the endianness right (reverse string in pairs of 2)
    activation_bytes = b"".join(reversed([activation_bytes[i:i+2] for i in
                                         range(0, len(activation_bytes), 2)]))
    activation_bytes = activation_bytes.decode("ascii")

    return activation_bytes, output


def fetch_activation_bytes(player_token, filename=None):

    base_url_license = "https://www.audible.com"
    rurl = base_url_license + "/license/licenseForCustomerToken"
    # register params
    params = {
        "customer_token": player_token
    }
    # deregister params
    dparams = {
        "customer_token": player_token,
        "action": "de-register"
    }

    headers = {
        "User-Agent": "Audible Download Manager"
    }

    with httpx.Client(headers=headers) as session:
        session.get(rurl, params=dparams)

        resp = session.get(rurl, params=params)
        register_response_content = resp.content

        if filename:
            pathlib.Path(filename).write_bytes(register_response_content)

        activation_bytes, _ = extract_activation_bytes(register_response_content)

        session.get(rurl, params=dparams)

        return activation_bytes


def get_activation_bytes(auth, filename=None):

    player_token = get_player_token(auth)
    activation_bytes = fetch_activation_bytes(player_token, filename)

    return activation_bytes



import base64
import hmac
import json
import logging
import os
import pathlib
import re
import struct
from hashlib import sha256
from typing import Dict, Tuple

from pbkdf2 import PBKDF2
from pyaes import AESModeOfOperationCBC, Encrypter, Decrypter

logger = logging.getLogger('audible.aescipher')

BLOCK_SIZE = 16


def aes_cbc_encrypt(
    key: bytes, iv: bytes, data: str, padding="default"
) -> bytes:
    encrypter = Encrypter(AESModeOfOperationCBC(key, iv), padding=padding)
    encrypted = encrypter.feed(data) + encrypter.feed()
    return encrypted


def aes_cbc_decrypt(
    key: bytes, iv: bytes, encrypted_data: bytes, padding="default"
) -> str:
    decrypter = Decrypter(AESModeOfOperationCBC(key, iv), padding=padding)
    decrypted = decrypter.feed(encrypted_data) + decrypter.feed()
    return decrypted.decode("utf-8")


def create_salt(salt_marker: bytes, kdf_iterations: int):
    header = (salt_marker + struct.pack('>H', kdf_iterations) + salt_marker)
    salt = os.urandom(BLOCK_SIZE - len(header))
    return header, salt


def pack_salt(header, unpacked_salt):
    return header + unpacked_salt


def unpack_salt(packed_salt, salt_marker):
    mlen = len(salt_marker)
    hlen = mlen * 2 + 2

    if not (
            packed_salt[:mlen] == salt_marker and
            packed_salt[mlen + 2:hlen] == salt_marker
    ):
        raise ValueError("Check salt_marker.")

    kdf_iterations = struct.unpack('>H', packed_salt[mlen:mlen + 2])[0]
    salt = packed_salt[hlen:]
    return salt, kdf_iterations


def derive_from_pbkdf2(
        password: str,
        *,
        key_size: int,
        salt: bytes,
        kdf_iterations: int,
        hashmod,
        mac
):
    kdf = PBKDF2(password, salt, min(kdf_iterations, 65535), hashmod, mac)
    return kdf.read(key_size)


class AESCipher:
    """Encrypt/Decrypt data using password to generate key. The encryption
    algorithm used is symmetric AES in cipher-block chaining (CBC) mode.
    
    ``key_size`` may be 16, 24 or 32 (default).
    The key is derived via the PBKDF2 key derivation function (KDF) from the
    password and a random salt of 16 bytes (the AES block size) minus the
    length of the salt header (see below).
    The hash function used by PBKDF2 is SHA256 per default. You can pass a
    different hash function module via the ``hashmod`` argument. The module
    must adhere to the Python API for Cryptographic Hash Functions (PEP 247).
    PBKDF2 uses a number of iterations of the hash function to derive the key,
    which can be set via the ``kdf_iterations` keyword argumeent. The default
    number is 1000 and the maximum 65535.
    The header and the salt are written to the first block of the encrypted
    output (bytes mode) or written as key/value pairs (dict mode). The header
    consist of the number of KDF iterations encoded as a big-endian word bytes
    wrapped by ``salt_marker`` on both sides. With the default value of
    ``salt_marker = b'$'``, the header size is thus 4 and the salt 12 bytes.
    The salt marker must be a byte string of 1-6 bytes length.
    The last block of the encrypted output is padded with up to 16 bytes, all
    having the value of the length of the padding.
    All values in dict mode are written as base64 encoded string.
    """

    def __init__(
            self,
            password: str,
            *,
            key_size: int = 32,
            salt_marker: bytes = b"$",
            kdf_iterations: int = 1000,
            hashmod=sha256,
            mac=hmac
    ) -> None:

        if not 1 <= len(salt_marker) <= 6:
            raise ValueError('The salt_marker must be one to six bytes long.')

        if not isinstance(salt_marker, bytes):
            raise TypeError('salt_marker must be a bytes instance.')

        if kdf_iterations >= 65536:
            raise ValueError('kdf_iterations must be <= 65535.')

        self.password = password
        self.key_size = key_size
        self.hashmod = hashmod
        self.mac = mac
        self.salt_marker = salt_marker
        self.kdf_iterations = kdf_iterations

    def _encrypt(self, data: str) -> Tuple[bytes, bytes, bytes]:
        header, salt = create_salt(self.salt_marker, self.kdf_iterations)
        key = derive_from_pbkdf2(
            password=self.password,
            key_size=self.key_size,
            salt=salt,
            kdf_iterations=self.kdf_iterations,
            hashmod=self.hashmod,
            mac=self.mac
        )
        iv = os.urandom(BLOCK_SIZE)
        encrypted_data = aes_cbc_encrypt(key, iv, data)
        return pack_salt(header, salt), iv, encrypted_data

    def _decrypt(self, salt: bytes, iv: bytes, encrypted_data: bytes) -> str:
        try:
            salt, kdf_iterations = unpack_salt(salt, self.salt_marker)
        except ValueError:
            kdf_iterations = self.kdf_iterations

        key = derive_from_pbkdf2(
            password=self.password,
            key_size=self.key_size,
            salt=salt,
            kdf_iterations=kdf_iterations,
            hashmod=self.hashmod,
            mac=self.mac
        )
        return aes_cbc_decrypt(key, iv, encrypted_data)

    def to_dict(self, data: str) -> Dict[str, str]:
        salt, iv, encrypted_data = self._encrypt(data)

        return {
            "salt": base64.b64encode(salt).decode("utf-8"),
            "iv": base64.b64encode(iv).decode("utf-8"),
            "ciphertext": base64.b64encode(encrypted_data).decode("utf-8"),
            "info": 'base64-encoded AES-CBC-256 of JSON object'
        }

    def from_dict(self, data: dict) -> str:
        salt = base64.b64decode(data["salt"])
        iv = base64.b64decode(data["iv"])
        encrypted_data = base64.b64decode(data["ciphertext"])
        return self._decrypt(salt, iv, encrypted_data)

    def to_bytes(self, data: str) -> bytes:
        salt, iv, encrypted_data = self._encrypt(data)
        return salt + iv + encrypted_data

    def from_bytes(self, data: bytes) -> str:
        bs = BLOCK_SIZE
        salt = data[:bs]
        iv = data[bs:2 * bs]
        encrypted_data = data[2 * bs:]
        return self._decrypt(salt, iv, encrypted_data)

    def to_file(
            self, data: str, filename: pathlib.Path, encryption="json",
            indent=4
    ):
        if encryption == "json":
            encrypted_dict = self.to_dict(data)
            data_json = json.dumps(encrypted_dict, indent=indent)
            filename.write_text(data_json)

        elif encryption == "bytes":
            encrypted_data = self.to_bytes(data)
            filename.write_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\"..")

    def from_file(self, filename: pathlib.Path, encryption="json"):
        if encryption == "json":
            encrypted_json = filename.read_text()
            encrypted_dict = json.loads(encrypted_json)
            return self.from_dict(encrypted_dict)

        elif encryption == "bytes":
            encrypted_data = filename.read_bytes()
            return self.from_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\".")


def detect_file_encryption(filename: pathlib.Path):
    file = filename.read_bytes()
    encryption = None

    try:
        file = json.loads(file)
        if "adp_token" in file:
            encryption = False
        elif "ciphertext" in file:
            encryption = "json"
    except UnicodeDecodeError:
        encryption = "bytes"

    return encryption


def remove_file_encryption(source, target, password, **kwargs):
    source_file = pathlib.Path(source)
    encryption = detect_file_encryption(source_file)

    if not encryption:
        raise Exception("file is not encrypted")

    crypter = AESCipher(password, **kwargs)
    decrypted = crypter.from_file(source_file, encryption=encryption)
    pathlib.Path(target).write_text(decrypted)


def _decrypt_voucher(device_serial_number, customer_id, device_type, asin, voucher):
    # https://github.com/mkb79/Audible/issues/3#issuecomment-705262614
    buf = (device_type + device_serial_number + customer_id + asin).encode("ascii")
    digest = sha256(buf).digest()
    key = digest[0:16]
    iv = digest[16:]

    # decrypt "voucher" using AES in CBC mode with no padding
    b64d_voucher = base64.b64decode(voucher)
    plaintext = aes_cbc_decrypt(key, iv, b64d_voucher, padding="none").rstrip("\x00")

    try:
        return json.loads(plaintext)
    except json.JSONDecodeError:
        fmt = r"^{\"key\":\"(?P<key>.*?)\",\"iv\":\"(?P<iv>.*?)\","
        match = re.match(fmt, plaintext)
        return match.groupdict()


def decrypt_voucher_from_licenserequest(auth, license_response):
    """Decrypt the voucher from licenserequest response

    :param auth: An instance of an `Authenticator` clsss
    :type auth: audible.FileAuthenticator, audible.LoginAuthenticator
    :param license_response: The response content from a successful http 
        `POST` request to api endpoint /1.0/content/{asin}/licenserequest
    :type license_response: dict
    :returns: The decryptes license voucher with needed key and iv
    :rtype: dict

    .. note::
        A device registration is needed to use the auth instance for a 
        license request and to obtain the needed device data
    
    """
    # device data
    device_info = auth.device_info
    device_serial_number = device_info["device_serial_number"]
    device_type = device_info["device_type"]

    # user data
    customer_id = auth.customer_info["user_id"]

    # book specific data
    asin = license_response["content_license"]["asin"]
    encrypted_voucher = license_response["content_license"]["license_response"]

    return _decrypt_voucher(
        device_serial_number=device_serial_number,
        customer_id=customer_id,
        device_type=device_type,
        asin=asin,
        voucher=encrypted_voucher
    )



import json
import logging
from typing import Union, Optional

import httpx
from httpx._models import URL

from .auth import LoginAuthenticator, FileAuthenticator
from .exceptions import (
    BadRequest, NotFoundError, NotResponding, NetworkError, ServerError,
    Unauthorized, UnexpectedError, RatelimitError, RequestError
)
from .localization import LOCALE_TEMPLATES, Locale
from .utils import test_convert

logger = logging.getLogger('audible.client')


class AudibleAPI:
    """
    .. deprecated:: v0.4.0
       Use :class:`Client` or :class:`AsyncClient` instead
    """

    REQUEST_LOG = '{method} {url} has received {text}, has returned {status}'

    def __init__(
            self,
            auth: Optional[
                Union[LoginAuthenticator, FileAuthenticator]] = None,
            session=None,
            is_async=False,
            **options
    ) -> None:
        self.auth = auth

        self.is_async = is_async
        self.session = (
            session or
            (httpx.AsyncClient() if is_async else httpx.Client())
        )
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        self.session.headers.update(headers)

        locale = test_convert("locale", options.get("locale", auth.locale))
        domain = options.get("domain", locale.domain)
        self.api_root_url = options.get("url", f"https://api.audible.{domain}")

        self.timeout = options.get('timeout', 10)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.aclose()

    def __repr__(self):
        return f"<AudibleAPI Client async={self.is_async}>"

    def close(self):
        if self.is_async:
            logger.warning(
                "Please use aclose() method to close a async client."
            )
        return self.session.close()

    async def aclose(self):
        await self.session.aclose()

    def switch_marketplace(self, locale):
        locale = test_convert("locale", locale)
        domain = locale.domain
        self.api_root_url = f"https://api.audible.{domain}"

    @property
    def marketplace(self):
        for value in LOCALE_TEMPLATES.values():
            domain = value["domain"]
            api_root_for_domain = f"https://api.audible.{domain}"

            if api_root_for_domain == self.api_root_url:
                return value["country_code"]

    def switch_user(self, auth: Union[LoginAuthenticator, FileAuthenticator]):
        self.auth = auth

    def get_user_profile(self):
        self.auth.refresh_access_token()
        return self.auth.user_profile()

    @property
    def user_name(self):
        user_profile = self.get_user_profile()
        return user_profile["name"]

    def _raise_for_status(self, resp, text, *, method=None, return_raw=True):
        try:
            data = json.loads(text)
        except json.JSONDecodeError:
            data = text
        code = getattr(resp, 'status', None) or getattr(resp, 'status_code')

        logger.debug(
            self.REQUEST_LOG.format(
                method=method or resp.request_info.method,
                url=resp.url,
                text=text,
                status=code
            )
        )

        if 300 > code >= 200:  # Request was successful
            if not return_raw:
                return data
            return data, resp  # value, response 
        elif code == 400:
            raise BadRequest(resp, data)
        elif code in (401, 403):  # Unauthorized request - Invalid credentials
            raise Unauthorized(resp, data)
        elif code == 404:  # not found
            raise NotFoundError(resp, data)
        elif code == 429:
            raise RatelimitError(resp, data)
        elif code == 503:  # Maintainence
            raise ServerError(resp, data)
        else:
            raise UnexpectedError(resp, data)

    async def _arequest(self, method, url, **kwargs):
        timeout = kwargs.pop('timeout', self.timeout)
        return_raw = kwargs.pop("return_raw", True)

        try:
            resp = await self.session.request(
                method, url, timeout=timeout, auth=self.auth, **kwargs
            )
            return self._raise_for_status(
                resp, resp.text, method=method, return_raw=return_raw
            )
        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        finally:
            try:
                await resp.aclose()
            except UnboundLocalError:
                pass

    def _request(self, method, url, **kwargs):
        if self.is_async:  # return a coroutine
            return self._arequest(method, url, **kwargs)
        timeout = kwargs.pop('timeout', self.timeout)
        return_raw = kwargs.pop("return_raw", True)

        try:
            resp = self.session.request(
                method, url, timeout=timeout, auth=self.auth, **kwargs
            )
            return self._raise_for_status(
                resp, resp.text, method=method, return_raw=return_raw
            )
        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        finally:
            try:
                resp.close()
            except UnboundLocalError:
                pass

    def _split_kwargs(self, **kwargs):
        protected_kwargs = [
            "method", "url", "params", "data", "json", "headers", "cookies",
            "files", "auth", "timeout", "allow_redirects", "proxies", "verify",
            "stream", "cert", "return_raw"
        ]
        params = kwargs.pop("params", {})
        for key in list(kwargs.keys()):
            if key not in protected_kwargs:
                params[key] = kwargs.pop(key)

        return params, kwargs

    def get(self, path, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("GET", url, params=params, **kwargs)

    def post(self, path, body, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("POST", url, params=params, json=body, **kwargs)

    def delete(self, path, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("DELETE", url, params=params, **kwargs)


def convert_response_content(resp):
    try:
        return resp.json()
    except json.JSONDecodeError:
        return resp.text

class Client:
    _API_URL_TEMP = "https://api.audible."
    _API_VERSION = "1.0"
    _SESSION = httpx.Client
    _REQUEST_LOG = '{method} {url} has received {text}, has returned {status}'

    def __init__(
            self,
            auth: Optional[
                Union[LoginAuthenticator, FileAuthenticator]] = None,
            country_code: Optional[str] = None,
            timeout: int = 10
    ):
        locale = Locale(country_code.lower()) if country_code else auth.locale
        api_url = self._API_URL_TEMP + locale.domain
        headers = {
            "Accept": "application/json",
            "Accept-Charset": "utf-8",
            "Content-Type": "application/json"
        }
        self.session = self._SESSION(
            headers=headers, timeout=timeout, base_url=api_url, auth=auth
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        return f"<Sync Client for *{self.marketplace}* marketplace>"

    def close(self):
        self.session.close()

    def switch_marketplace(self, country_code):
        locale = Locale(country_code.lower())
        self.session.base_url = URL(self._API_URL_TEMP + locale.domain)

    @property
    def marketplace(self):
        base_url = str(self.session.base_url)
        slice_len = len(self._API_URL_TEMP)
        domain = base_url[slice_len:]

        for country in LOCALE_TEMPLATES.values():
            if domain == country["domain"]:
                return country["country_code"]

    @property
    def auth(self):
        return self.session.auth

    def switch_user(
            self,
            auth: Union[LoginAuthenticator, FileAuthenticator],
            switch_to_default_marketplace: bool = False
    ):
        if switch_to_default_marketplace:
            self.switch_marketplace(auth.locale.country_code)
        self.session.auth = auth

    def get_user_profile(self):
        self.auth.refresh_access_token()
        return self.auth.user_profile()

    @property
    def user_name(self):
        user_profile = self.get_user_profile()
        return user_profile["name"]

    def _raise_for_status_error(self, resp):
        code = resp.status_code

        data = convert_response_content(resp)

        if 300 > code >= 200:  # Request was successful
            return
        elif code == 400:
            raise BadRequest(resp, data)
        elif code in (401, 403):  # Unauthorized request - Invalid credentials
            raise Unauthorized(resp, data)
        elif code == 404:  # not found
            raise NotFoundError(resp, data)
        elif code == 429:
            raise RatelimitError(resp, data)
        elif code == 503:  # Maintainence
            raise ServerError(resp, data)
        else:
            raise UnexpectedError(resp, data)

    def _prepare_path(self, path):
        if path.startswith("/"):
            path = path[1:]

        if path.startswith(self._API_VERSION):
            return path

        return "/".join((self._API_VERSION, path))

    def _request(self, method: str, path: str, **kwargs):
        url = self._prepare_path(path)

        try:
            resp = self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method,
                    url=resp.url,
                    text=resp.text,
                    status=resp.status_code
                )
            )

            resp.raise_for_status()

            return convert_response_content(resp)

        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        except httpx.RequestError as exc:
            raise RequestError(exc)
        except httpx.HTTPStatusError as exc:
            self._raise_for_status_error(exc.response)

        finally:
            try:
                resp.close()
            except UnboundLocalError:
                pass

    def _split_kwargs(self, **kwargs):
        protected_kwargs = [
            "method", "url", "params", "data", "json", "headers", "cookies",
            "files", "auth", "timeout", "allow_redirects", "proxies", "verify",
            "stream", "cert"
        ]
        params = kwargs.pop("params", {})
        for key in list(kwargs.keys()):
            if key not in protected_kwargs:
                params[key] = kwargs.pop(key)

        return params, kwargs

    def get(self, path, **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("GET", path, params=params, **kwargs)

    def post(self, path, body, **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("POST", path, params=params, json=body, **kwargs)

    def delete(self, path, **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("DELETE", path, params=params, **kwargs)


class AsyncClient(Client):
    _SESSION = httpx.AsyncClient

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    def __repr__(self):
        return f"<AyncClient for *{self.marketplace}* marketplace>"

    async def close(self):
        await self.session.aclose()

    async def _request(self, method: str, path: str, **kwargs):
        url = self._prepare_path(path)

        try:
            resp = await self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method,
                    url=resp.url,
                    text=resp.text,
                    status=resp.status_code
                )
            )

            resp.raise_for_status()

            return convert_response_content(resp)

        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        except httpx.RequestError as exc:
            raise RequestError(exc)
        except httpx.HTTPStatusError as exc:
            self._raise_for_status_error(exc.response)

        finally:
            try:
                await resp.aclose()
            except UnboundLocalError:
                pass



# -*- coding: utf-8 -*-

from .client import AudibleAPI, Client, AsyncClient
from .auth import LoginAuthenticator, FileAuthenticator
from ._logging import AudibleLogHelper
from audible._version import __version__


__all__ = [
    "__version__", "AudibleAPI", "LoginAuthenticator", "FileAuthenticator",
    "log_helper", "Client", "AsyncClient"
]


log_helper = AudibleLogHelper()



import logging
import pathlib
from typing import Optional, Union
from warnings import warn

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "not-set": logging.NOTSET
}


class AudibleLogHelper:
    def __init__(self) -> None:
        self._formatter = logging.Formatter(
            "%(asctime)s %(levelname)s [%(name)s] "
            "%(filename)s:%(lineno)d: %(message)s"
        )
        self._logger = logging.getLogger("audible")

    def set_level(self, level: Union[str, int]) -> None:
        """Set logging level for the main logger."""
        if isinstance(level, str):
            level = LEVELS.get(level.lower().strip())

        self._logger.setLevel(level)
        self._logger.info(
            "set logging threshold to \"%s\"",
            logging.getLevelName(self._logger.level)
        )

    def _set_handler_level(self, handler,
                           level: Optional[Union[str, int]]) -> None:
        if isinstance(level, str):
            level = LEVELS.get(level.lower().strip())

        if level:
            handler.setLevel(level)

        self._logger.info(
            f"set logging threshold for \"{handler.name}\" "
            f"to \"{logging.getLevelName(handler.level)}\""
        )

        if handler.level < self._logger.level:
            warn("Handler level must be equal or greater than logger level")

    def set_console_logger(self,
                           level: Optional[Union[str, int]] = None) -> None:
        """Set logging level for the stream handler."""
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(self._formatter)
        stream_handler.set_name("ConsoleLogger")
        self._logger.addHandler(stream_handler)
        self._set_handler_level(stream_handler, level)

    def set_file_logger(
            self, filename: str, level: Optional[Union[str, int]] = None
    ) -> None:
        """Set logging level and filename for the file handler."""
        filename = pathlib.Path(filename)
        file_handler = logging.FileHandler(filename)
        file_handler.setFormatter(self._formatter)
        file_handler.set_name("FileLogger")
        self._logger.addHandler(file_handler)
        self._set_handler_level(file_handler, level)

    def capture_warnings(self, status: bool = True) -> None:
        logging.captureWarnings(status)
        self._logger.info(
            f"Capture warnings {'activated' if status else 'deactivated'}"
        )



import base64
import json
import logging
from collections.abc import MutableMapping
from datetime import datetime, timedelta
from typing import Any, Dict

import httpx
import rsa

from .activation_bytes import get_activation_bytes as get_ab
from .aescipher import AESCipher, detect_file_encryption
from .exceptions import FileEncryptionError, NoAuthFlow, NoRefreshToken
from .login import login
from .register import deregister as deregister_
from .register import register as register_
from .utils import test_convert

logger = logging.getLogger('audible.auth')


def refresh_access_token(refresh_token: str, domain: str) -> Dict[str, Any]:
    """
    Refresh access token with refresh token.
    Access tokens are valid for 60 mins.
    
    """

    body = {
        "app_name": "Audible",
        "app_version": "3.26.1",
        "source_token": refresh_token,
        "requested_token_type": "access_token",
        "source_token_type": "refresh_token"
    }

    resp = httpx.post(f"https://api.amazon.{domain}/auth/token", data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    expires_in_sec = int(resp_dict["expires_in"])
    expires = (datetime.utcnow() + timedelta(seconds=expires_in_sec
                                             )).timestamp()

    return {"access_token": resp_dict["access_token"], "expires": expires}


def refresh_website_cookies(
        refresh_token: str, domain: str, cookies_domain: str
) -> Dict[str, str]:
    url = f"https://www.amazon.{domain}/ap/exchangetoken"

    body = {
        "app_name": "Audible",
        "app_version": "3.26.1",
        "source_token": refresh_token,
        "requested_token_type": "auth_cookies",
        "source_token_type": "refresh_token",
        "domain": f".amazon.{cookies_domain}"
    }

    resp = httpx.post(url, data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    raw_cookies = resp_dict["response"]["tokens"]["cookies"]
    website_cookies = dict()
    for cookies_domain in raw_cookies:
        for cookie in raw_cookies[cookies_domain]:
            website_cookies[cookie["Name"]] = cookie["Value"].replace(r'"',
                                                                      r'')

    return website_cookies


def user_profile(access_token: str, domain: str) -> Dict[str, Any]:
    """Returns user profile from amazon."""

    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(
        f"https://api.amazon.{domain}/user/profile", headers=headers
    )
    resp.raise_for_status()

    return resp.json()


def user_profile_audible(access_token: str, domain: str) -> Dict[str, Any]:
    """Returns user profile from audible."""

    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(
        f"https://api.audible.{domain}/user/profile", headers=headers
    )
    resp.raise_for_status()

    return resp.json()


def sign_request(
        method: str, path: str, body: bytes, adp_token: str, private_key: str
) -> Dict[str, str]:
    """
    Helper function who creates a signed header for requests.

    :param path: the requested url path and query
    :param method: the request method (GET, POST, DELETE, ...)
    :param body: the http message body
    :param adp_token: the token is obtained after register as device
    :param private_key: the rsa key obtained after register as device
    """
    date = datetime.utcnow().isoformat("T") + "Z"
    body = body.decode("utf-8")

    data = f"{method}\n{path}\n{date}\n{body}\n{adp_token}"

    key = rsa.PrivateKey.load_pkcs1(private_key)
    cipher = rsa.pkcs1.sign(data.encode(), key, "SHA-256")
    signed_encoded = base64.b64encode(cipher)

    signature = f"{signed_encoded.decode()}:{date}"

    return {
        "x-adp-token": adp_token,
        "x-adp-alg": "SHA256withRSA:1.0",
        "x-adp-signature": signature
    }


class BaseAuthenticator(MutableMapping, httpx.Auth):
    """Base Class for retrieve and handle credentials."""

    requires_request_body = True

    def __getitem__(self, key):
        return self.__dict__[key]

    def __getattr__(self, attr):
        try:
            return self.__getitem__(attr)
        except KeyError:
            return None

    def __delitem__(self, key):
        del self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = test_convert(key, value)

    def __setattr__(self, attr, value):
        self.__setitem__(attr, value)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return f"{type(self).__name__}({self.__dict__})"

    def auth_flow(self, request: httpx.Request):
        if self.adp_token and self.device_private_key:
            logger.info("Use sign request auth flow.")
            self.sign_request(request)

        elif self.access_token:
            if self.access_token_expired:
                self.refresh_access_token()
            logger.info("Use access token auth flow.")
            headers = {
                "Authorization": "Bearer " + self.access_token,
                "client-id": "0"
            }
            request.headers.update(headers)

        else:
            message = "No auth flow method available."
            logger.critical(message)
            raise NoAuthFlow(message)

        yield request

    def sign_request(self, request: httpx.Request):
        method = request.method
        path = request.url.path
        query = request.url.query
        body = request.content

        if query:
            path += f"?{query}"

        headers = sign_request(
            method, path, body, self.adp_token, self.device_private_key
        )

        request.headers.update(headers)

    def to_file(
            self,
            filename=None,
            password=None,
            encryption="default",
            indent=4,
            set_default=True,
            **kwargs
    ):

        if not (filename or self.filename):
            raise ValueError("No filename provided")

        if filename:
            filename = test_convert("filename", filename)

        target_file = filename or self.filename

        if encryption != "default":
            encryption = test_convert("encryption", encryption)
        else:
            encryption = self.encryption or False

        body = {
            "website_cookies": self.website_cookies,
            "adp_token": self.adp_token,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "device_private_key": self.device_private_key,
            "store_authentication_cookie": self.store_authentication_cookie,
            "device_info": self.device_info,
            "customer_info": self.customer_info,
            "expires": self.expires,
            "locale_code": self.locale.country_code
        }
        json_body = json.dumps(body, indent=indent)

        if encryption is False:
            target_file.write_text(json_body)
            crypter = None
        else:
            if password:
                crypter = test_convert("crypter",
                                       AESCipher(password, **kwargs))
            elif self.crypter:
                crypter = self.crypter
            else:
                raise ValueError("No password provided")

            crypter.to_file(
                json_body,
                filename=target_file,
                encryption=encryption,
                indent=indent
            )

        logger.info(f"saved data to file {filename}")

        if set_default:
            self.filename = target_file
            self.encryption = encryption
            self.crypter = crypter

        logger.info(f"set filename {target_file} as default")

    def re_login(
            self,
            username,
            password,
            captcha_callback=None,
            otp_callback=None,
            cvf_callback=None
    ):

        login_device = login(
            username=username,
            password=password,
            country_code=self.locale.country_code,
            domain=self.locale.domain,
            market_place_id=self.locale.market_place_id,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback,
            cvf_callback=cvf_callback
        )

        self.update(**login_device)

    def register_device(self):
        register_device = register_(
            access_token=self.access_token, domain=self.locale.domain
        )

        self.update(**register_device)

    def deregister_device(self, deregister_all: bool = False):
        return deregister_(
            access_token=self.access_token,
            deregister_all=deregister_all,
            domain=self.locale.domain
        )

    def refresh_access_token(self, force=False):
        if force or self.access_token_expired:
            if self.refresh_token is None:
                message = "No refresh token found. Can't refresh access token."
                logger.critical(message)
                raise NoRefreshToken(message)
            refresh_data = refresh_access_token(
                refresh_token=self.refresh_token, domain=self.locale.domain
            )

            self.update(**refresh_data)
        else:
            logger.info(
                "Access Token not expired. No refresh nessessary. "
                "To force refresh please use force=True"
            )

    def get_activation_bytes(self, filename=None):
        return get_ab(self, filename)

    def user_profile(self):
        return user_profile(
            access_token=self.access_token, domain=self.locale.domain
        )

    @property
    def access_token_expires(self):
        return datetime.fromtimestamp(self.expires) - datetime.utcnow()

    @property
    def access_token_expired(self):
        return datetime.fromtimestamp(self.expires) <= datetime.utcnow()


class LoginAuthenticator(BaseAuthenticator):
    """Authenticator class to retrieve credentials from login."""

    def __init__(
            self,
            username: str,
            password: str,
            locale,
            register=False,
            captcha_callback=None,
            otp_callback=None,
            cvf_callback=None
    ):
        self.locale = locale

        resp = login(
            username=username,
            password=password,
            country_code=self.locale.country_code,
            domain=self.locale.domain,
            market_place_id=self.locale.market_place_id,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback,
            cvf_callback=cvf_callback
        )

        logger.info(f"logged in to audible as {username}")

        if register:
            resp = register_(resp["access_token"], self.locale.domain)
        logger.info("registered audible device")

        self.update(**resp)


class FileAuthenticator(BaseAuthenticator):
    """Authenticator class to retrieve credentials from stored file."""

    def __init__(
            self, filename, password=None, locale=None, encryption=None,
            **kwargs
    ) -> None:

        self.filename = filename
        self.encryption = encryption or detect_file_encryption(self.filename)

        if self.encryption:
            if password is None:
                message = "File is encrypted but no password provided."
                logger.critical(message)
                raise FileEncryptionError(message)
            self.crypter = AESCipher(password, **kwargs)
            file_data = self.crypter.from_file(self.filename, self.encryption)
        else:
            file_data = self.filename.read_text()

        json_data = json.loads(file_data)

        locale_code = json_data.pop("locale_code", None)
        locale = locale or locale_code
        self.locale = locale

        # login cookies where renamed to website cookies
        # old names must be adjusted
        login_cookies = json_data.pop("login_cookies", None)
        if login_cookies:
            json_data["website_cookies"] = login_cookies

        self.update(**json_data)

        logger.info(
            (
                f"load data from file {self.filename} for "
                f"locale {self.locale.country_code}"
            )
        )
